/*
 * ** header with license infoc
 * ...
 */

'use strict';

/**
 * This module adds to forms to LinkAhead that relate to sample
 * management; there are forms for sample registration, upload, and
 * export.
 *
 * @module ext_flight_preview
 * @version 0.1
 *
 * @requires nothing yet
 * (pass the dependencies as arguments)
 */
const ext_samplemanagement = function($, navbar, log, form_elements, form_panel, sample_management_reference_resolver) {
    const logger = log.getLogger("samplemanagement_form");
    const tool_box = "Sample Management"
    const upload_sample_template_title = "Upload sample template"; // title of the form and text in the toolbox
    const upload_sample_template_panel_id = "upload_sample_template_form_panel";
    const register_new_samples_title = "Register new samples"; // title of the form and text in the toolbox
    const register_new_samples_panel_id = "register_new_samples_form_panel";
    const registerNewChildSamplesTitle = "Register new child samples"; // title of the form and text in the toolbox
    const registerNewChildSamplesPanelId = "register_new_children_form_panel";
    const upload_pdf_id = "upload_pdf_form_panel";
    const upload_pdf_title = "Upload PDF file";
    const export_csv_id = "export_sample_csv";
    const export_csv_title = "Export Samples to CSV";
    const sample_id_label = ("${BUILD_MODULE_EXT_SAMPLEMANGEMENT_ID_LABEL}" != "") ? "${BUILD_MODULE_EXT_SAMPLEMANGEMENT_ID_LABEL}" : "Sample ID";
    const accepted_file_formats = [
        ".csv",
        ".CSV",
        "application/csv",
        "text/csv",
    ];
    const required_column_names = [
        "LinkAhead ID",
        "Main User",
        "Latitude start",
        "Longitude start",
        "Elevation start",
        "Storage ID",
        "Device"
    ];
    const requiredColumnNamesChildren = [
        "LinkAhead ID",
        "Main User",
        "Parent LinkAhead ID",
        "SampleType",
        "Storage ID",
    ];
    const non_sample_rt_column_names = [
        "Biome",
        "Campaign",
        "Start date",
        "End date",
        "IGSN DOI",
        "Level",
        "Locality description",
        "Locality name",
        "Sphere",
        "Event responsible",
        "Latitude stop",
        "Longitude stop",
        "Elevation stop",
        "PDFReport",
        "Parent LinkAhead ID",
    ];
    const allColumnNames = non_sample_rt_column_names.concat(requiredColumnNamesChildren).concat(required_column_names);

    const unused_property_names = [
        "Main User",
        'NagoyaCase',
        "Parent_Sample",
    ]

    const upload_sample_template_form_config = {
        script: "upload_sample_template.py",
        fields: [{
            type: "file",
            name: "sample_template_file",
            label: "Sample import template",
            required: true,
            cached: false,
            accept: accepted_file_formats.join(","),
            help: "Select the filled out import template you want to upload."
        }, ],
    };

    const export_csv_form_config = {
        script: "export_sample_csv.py",
        fields: [{
            type: "text",
            name: "bis_ids",
            label: sample_id_label,
            required: false,
            pattern: "((([0-9]+\\s*-\\s*){0,1}[0-9]+)\\s*,\\s*)*(([0-9]+\\s*-\\s*){0,1}[0-9]+)",
            cached: false,
            help: `Comma-separated list of single ${sample_id_label}s or ranges, e.g. '124, 126-139, 242'`,
        }, {
            type: "text",
            name: "query_string",
            label: "Query string",
            help: "As an alternative, specify the samples by pasting a valid select or find query in this field. All samples in the result set will be exported."
        }, {
            type: "checkbox",
            name: "noEmpyColumns",
            label: "Remove empty columns",
            help: "Columns that are empty for all exported samples will be deleted from the exported CSV file."
        }],
    };

    /**
     * Reset the column names select field of the register samples form with
     * the values of the template selected in the template field
     */
    const reset_columns_field = async function(form) {
        var template_field = form_elements.get_fields(form, "register_sample_template")[0];
        var template_id = $(template_field).find("select:input")[0].value;
        var template_ent = await transaction.retrieveEntityById(template_id);
        var names = $(template_ent).find("Property[name='column_names']").find("Value").toArray().map(e => e.textContent);
        var col_field = form_elements.get_fields(form, "column_names")[0];
        $(col_field).find(".selectpicker").selectpicker('val', names)
    };

    /**
     * The register sample form has some special interaction among fields.
     * This function create the form and adds callbacks that facilitate thiese
     * interactions.
     *
     * @return {HTMLElement} form
     */
    const _initRegisterSampleForm = function(requiredColumnNames, showTemplateField) {
        const register_new_samples_form_config = {
            script: "register_new_samples.py",
            fields: [{
                    type: "reference_drop_down",
                    name: "responsible_person",
                    label: "Main user",
                    query: "FIND RECORD Person",
                    make_desc: sample_management_reference_resolver.resolve_person_reference,
                    required: true
                },
                {
                    type: "integer",
                    name: "number_of_samples",
                    label: "Number of Samples",
                    required: true
                },
                //{ type: "checkbox", name: "samples_from_cruise", label: "Samples from a cruise", required: false, help: "When this option is selected the template will include fields for cruise-secific sampling, e.g. the platform, campaign, station and hol."},
                //{ type: "checkbox", name: "start_and_end_time", label: "Include start and end times", required: false, help: "If this option is selected, the template will include start- and end-field for Date and Time. Else it will only include one date and one time field."},
                //{ type: "integer", name: "number_of_locations", label: "Positions per Event", required: true, help: "The number of geographic positions you want to include for each sampling even. At least one position is required. If the sampling is stationary one should be enought. But for example, if you are trawling, you can include multiple positions of the trawl."},
                {
                    type: "select",
                    name: "column_names",
                    label: "Column names:",
                    options: [],
                    required: false,
                    multiple: true
                },
            ],
        };
        if (showTemplateField) {
            register_new_samples_form_config.fields = [{
                type: "reference_drop_down",
                name: "register_sample_template",
                label: "Template",
                query: "FIND RECORD RegisterSampleTemplate",
                required: false,
                help: "Templates allow to store frequently used sets of column names.",
                make_desc: getEntityName
            }].concat(register_new_samples_form_config.fields);
        }

        var form = form_elements.make_form(register_new_samples_form_config);
        if (showTemplateField) {
            // add callback that resets column names field when template changes

            form_elements.field_ready(
                    form_elements.get_fields(form, "register_sample_template")[0])
                .then((field) => ($(field).find(":input").change(() => reset_columns_field(form))))
                .catch(err => logger.error(err));
        }

        // refresh the options of the column names select field once the list
        // is available (The sample RT needs to be retrieved
        var column_name_select = $(form).find("select[name='column_names']");
        collect_column_names(requiredColumnNames).then(function(names) {
            for (let option of names) {
                column_name_select.append(
                    form_elements._make_option(option.value, option.label));
            }
            column_name_select.selectpicker('refresh');
        }).catch(err => logger.error(err));
        var col_row = $(form).find(".row[data-field-name='column_names']")[0];
        col_row.before($(
            '<div class="row caosdb-f-field caosdb-v-field caosdb-f-form-field-required" data-field-name="required_column_names"><label for="required_column_names" data-property-name="required_column_names" class="col-form-label col-sm-3">Required Columns</label><div class="caosdb-f-property-value col-sm-9"><textarea class="form-control rows="2"  name="required_column_names" required="" disabled>' + requiredColumnNames.join(", ") + '</textarea></div></div>')[0]);
        return form
    }

    const initRegisterSampleForm = function() {
        return _initRegisterSampleForm(required_column_names, true);
    }

    const initRegisterChildSamplesForm = function() {
        return _initRegisterSampleForm(requiredColumnNamesChildren, false);
    }

    const initRegisterTemplateForm = function() {
        const config = {
            script: "register_sample_template.py",
            fields: [{
                type: "text",
                name: "template_name",
                label: "Name of the new sample-registration template",
                required: true
            }, {
                type: "reference_drop_down",
                name: "main_user",
                label: "Main user",
                query: "FIND RECORD Person",
                make_desc: sample_management_reference_resolver.resolve_person_reference,
                required: true
            }, {
                type: "select",
                name: "column_names",
                label: "Column names",
                options: [],
                required: true,
                multiple: true
            }]
        };
        const form = form_elements.make_form(config);

        // TODO(fspreck): Tidy up to remove code duplications.
        const column_name_select = $(form).find("select[name='column_names']");
        collect_column_names(required_column_names).then(names => {
            for (let option of names) {
                column_name_select.append(
                    form_elements._make_option(option.value, option.label)
                );
            }
            column_name_select.selectpicker("refresh");
        }).catch(err => logger.error(err));
        var col_row = $(form).find(".row[data-field-name='column_names']")[0];
        col_row.before($(
            '<div class="row caosdb-f-field caosdb-v-field caosdb-f-form-field-required" data-field-name="required_column_names"><label for="required_column_names" data-property-name="required_column_names" class="col-form-label col-sm-3">Required Columns (always present)</label><div class="caosdb-f-property-value col-sm-9"><textarea class="form-control rows="2"  name="required_column_names" required="" disabled>' + required_column_names.join(", ") + '</textarea></div></div>')[0]);
        return form
    }

    /**
     * Creates a list of possible column names for a fixed list and the
     * properties of the sample RT
     *
     * @return {Array} Options
     */
    const collect_column_names = async function(requiredColumnNames) {
        const a = await query('find recordtype with name=sample');
        const column_names = $(a[0]).find(".caosdb-property-name").toArray().map(e => e.textContent);
        var options = [...new Set(allColumnNames.concat(column_names))];
        options = options.filter(n => !(unused_property_names.includes(n) || requiredColumnNames.includes(n))).sort();
        return options.map(e => ({
            value: e,
            label: e
        }));
    }

    const get_pdf_record_type_id = async (name) => {
        const response = await connection.get("Entity/?query=SELECT id FROM RECORDTYPE " + name);
        return response.evaluate("/Response/Entity/@id", response, null, XPathResult.STRING_TYPE)?.stringValue;
    }

    const init_pdf_upload = () => {
        const container = $(`<div class="row"/>`);
        const config = {
            directoryBase: "samples/pdfs",
            directoryReadOnly: true,
            directoryShow: false,
            recordTypeShow: false,
            recordType: get_pdf_record_type_id("PDFReport"),
            accept: "application/pdf,.pdf",
            onCancel: (e) => {
                // e.target.PARENTNODE.form here, because thats how the file
                // upload library works
                e.target.parentNode.form.dispatchEvent(form_elements.cancel_form_event);
            },
            onFinish: (e) => {
                // e.target.form here, because thats how our own buttons work
                e.target.form.dispatchEvent(form_elements.cancel_form_event);
            },
        }
        caosdb_file_upload_widget.create_file_upload_widget(container[0], config);
        return container[0];
    }

    /**
     * Add a button to the navbar, saying "Upload CSV File" which opens a
     * form for file upload.
     */
    const init_show_samplemanagement_panel_button = async function() {
        //var form_wrapper = form_elements.make_form(config);
        navbar.add_tool(upload_sample_template_title, tool_box, {
            callback: form_panel.create_show_form_callback(
                upload_sample_template_panel_id,
                upload_sample_template_title,
                upload_sample_template_form_config)
        });
        navbar.add_tool(register_new_samples_title, tool_box, {
            callback: form_panel.create_show_form_callback(
                register_new_samples_panel_id,
                register_new_samples_title,
                undefined,
                initRegisterSampleForm)
        });
        navbar.add_tool(registerNewChildSamplesTitle, tool_box, {
            callback: form_panel.create_show_form_callback(
                registerNewChildSamplesPanelId,
                registerNewChildSamplesTitle,
                undefined,
                initRegisterChildSamplesForm)
        });
        navbar.add_tool("Create sample template", tool_box, {
            callback: form_panel.create_show_form_callback(
                "create-sample-template",
                "Create sample template",
                undefined,
                initRegisterTemplateForm)
        });
        // Set auto_focus=false because of WebUI bug: https://gitlab.com/linkahead/linkahead-webui/-/issues/258
        navbar.add_tool(upload_pdf_title, tool_box, {
            callback: form_panel.create_show_form_callback(
                upload_pdf_id,
                upload_pdf_title,
                undefined,
                init_pdf_upload,
                false
            )
        });
        const csv_callback = form_panel.create_show_form_callback(
            export_csv_id,
            export_csv_title,
            export_csv_form_config
        );
        navbar.add_tool(export_csv_title, tool_box, {
            callback: (e) => {
                csv_callback(e);
                const input_field = document.querySelector(`#${export_csv_id} form`)[0]

                const check_pattern = () => {
                    input_field.checkValidity();
                    input_field.reportValidity();
                }
                input_field.addEventListener("blur", check_pattern);
            }
        });
    };


    const is_authorized = function() {
        // return true //for testing
        return isAuthenticated() //&& (userHasRole("stockmanager") || userHasRole("administration"))
    }


    const init = function() {
        if (is_authorized()) {
            init_show_samplemanagement_panel_button();
        }
    }


    /* the main function must return the initialization of the module */
    return {
        init: init,
    };
}($, navbar, log, form_elements, form_panel, sample_management_reference_resolver);

// this will be replaced by require.js in the future.
$(document).ready(function() {
    // use a variable starting with `BUILD_MODULE_` to enable your module
    // the build variable has to be enabled in the `build.properties.d/` directory.
    // Otherwise the module will not be activated.
    if ("${BUILD_MODULE_EXT_SAMPLEMANAGEMENT}" === "ENABLED") {
        caosdb_modules.register(ext_samplemanagement);
    }
});
